﻿using UnityEngine;
using System.Collections;

namespace Monolizm
{
	/// <summary>
	/// フェイスアニメーション管理クラス.
	/// </summary>
	public class FaceAnimationController : MonoBehaviour
	{

		#region public enumerate -----------------------------------------------------------------
		/// <summary>
		/// 目の種類.
		/// </summary>
		public enum EyeTypes:int
		{
			Default,		/// デフォルト.
			Angry,			/// 怒ってる.
			Blink,			/// 瞬き.
			Close,			/// 閉じている.
			Open,			/// 見開いている.
			Dizzy,			/// フラフラ.
			Happy,			/// 嬉しい.
			Hmm,			/// ふーん.
			Feel,			/// 感じる.
			Wink,			/// ウィンク.
			Dot,			/// 目が点.
			Wet,			/// 潤んでる.
			Surprised,		/// 驚き.
			Boast,			/// 自慢.
			NUM,			/// 目の種類数.
		}


		/// <summary>
		/// 口の種類.
		/// </summary>
		public enum MouthTypes:int
		{
			Default,		/// デフォルト.
			Smile,			/// 笑顔.
			A,				/// あ.
			I,				/// い.
			U,				/// う.
			E,				/// え.
			O,				/// お.
			Happy,			/// 嬉しい.
			Nmm,			/// ん.
			Chu,			/// チュ.
			NUM,			/// 口の種類数.
		}
		#endregion



		#region public variable ------------------------------------------------------------------
		public	Material			EyeMaterial;								/// 目のマテリアル
		public	Material			MouthMaterial;								/// 口のマテリアル
		public	GameObject			LeftEye;									/// 左目のゲームオブジェクト
		public	GameObject			RightEye;									/// 右目のゲームオブジェクト
		#endregion



		#region private variable -----------------------------------------------------------------
		/// <summary>
		/// アニメーション情報
		/// </summary>
		private struct AnimationInfo
		{
			public	Vector2			TextureSize;								/// テクスチャのサイズ.
			public	Rect			Atlas;										/// アトラス情婦.
			public	int				Type;										/// 種類.
			public	Material		Mat;										/// マテリアル.
			public	int				VNum;										/// 縦方向のアトラス数
			public	int				HNum;										/// 横方向のアトラス数
		}
		private	AnimationInfo		m_eyeInfo;									/// 目のアニメーション情報
		private	AnimationInfo		m_mouthInfo;								/// 目のアニメーション情報


		/// <summary>
		/// 瞬きの状態
		/// </summary>
		private enum BlinkState 
		{
			READY,				/// 目を開いている状態
			CLOSING,			/// 目を閉じている途中
			OPENING,			/// 目を開いている途中
		}

		/// <summary>
		/// 目の種類
		/// </summary>
		private enum EYES
		{
			LEFT,				/// 左目
			RIGHT,				/// 右目
			NUM,				/// 目の数
		}

		/// <summary>
		/// 瞬き制御
		/// </summary>
		private struct Blink
		{
			public BlinkState		State;										/// 瞬きの状態
			public float			MotionTimer;								/// 動き制御タイマー
			public float			Speed;										/// 瞬きのスピード
			public float			IntervalTimer;								/// 瞬きのtタイミング管理タイマー
			public float			Interval;									/// 瞬きのタイミング
			public float			IntervalMin;								/// 瞬きのタイミング最小値
			public float			IntervalMax;								/// 瞬きのタイミング最大値
		}
		private Blink[]				m_blinkEye = new Blink[(int)EYES.NUM];		/// 瞬き制御
		private SkinnedMeshRenderer m_leftEyeSkinnedMeshRenderer;				/// 左目瞬き用ブレンドシェイプ
		private SkinnedMeshRenderer m_rightEyeSkinnedMeshRenderer;				/// 右目瞬き用ブレンドシェイプ


		[SerializeField]
		private	float				BlinkSpeed;									/// 瞬きの速さ(瞬きするのにどれくらいの時間をかけるか:秒)
		[SerializeField]
		private	float				BlinkIntervalMin;							/// 瞬きを発動させる間隔最小値:秒
		[SerializeField]
		private	float				BlinkIntervalMax;							/// 瞬きを発動させる間隔最大値:秒


		[SerializeField]
		private	Vector2				EyeAtlasSize;								/// 目テクスチャ内の一つのマップサイズ.
		[SerializeField]
		private	Vector2				MouthAtlasSize;								/// 口テクスチャ内の一つのマップサイズ.
		#endregion



		#region Unity MonoBehaviour OverRide------------------------------------------------------
		void Awake ()
		{
			// 目のテクスチャ情報
			m_eyeInfo.Mat			= EyeMaterial;
			Texture texture			= m_eyeInfo.Mat.mainTexture;
			m_eyeInfo.TextureSize.x = texture.width;
			m_eyeInfo.TextureSize.y = texture.height;
			m_eyeInfo.Atlas.width	= EyeAtlasSize.x;
			m_eyeInfo.Atlas.height	= EyeAtlasSize.y;
			m_eyeInfo.VNum 			= (int)(m_eyeInfo.TextureSize.y / m_eyeInfo.Atlas.height);
			m_eyeInfo.HNum 			= (int)(m_eyeInfo.TextureSize.x / m_eyeInfo.Atlas.width);

			// 口のテクスチャ情報
			m_mouthInfo.Mat				= MouthMaterial;
			texture						= m_mouthInfo.Mat.mainTexture;
			m_mouthInfo.TextureSize.x 	= texture.width;
			m_mouthInfo.TextureSize.y 	= texture.height;
			m_mouthInfo.Atlas.width		= MouthAtlasSize.x;
			m_mouthInfo.Atlas.height	= MouthAtlasSize.y;
			m_mouthInfo.VNum 			= (int)(m_mouthInfo.TextureSize.y / m_mouthInfo.Atlas.height);
			m_mouthInfo.HNum 			= (int)(m_mouthInfo.TextureSize.x / m_mouthInfo.Atlas.width);


			// 目のスキンメッシュレンダラ
			m_leftEyeSkinnedMeshRenderer = LeftEye.GetComponent<SkinnedMeshRenderer>();
			m_rightEyeSkinnedMeshRenderer = RightEye.GetComponent<SkinnedMeshRenderer>();

		}

		void Start()
		{
			for (int i = 0; i < (int)EYES.NUM; i++)
			{
				m_blinkEye [i].State 			= BlinkState.READY;
				m_blinkEye [i].MotionTimer		= 0;
				m_blinkEye [i].Speed			= BlinkSpeed;
				m_blinkEye [i].IntervalTimer	= 0;
				m_blinkEye [i].IntervalMin		= BlinkIntervalMin;
				m_blinkEye [i].IntervalMax 		= BlinkIntervalMax;
				if (i == 0)
				{
					m_blinkEye [i].Interval = Random.Range (m_blinkEye [i].IntervalMin, m_blinkEye [i].IntervalMax);
				}
				else
				{
					// 両目で瞬きタイミングがずれないようにする
					m_blinkEye [i].Interval = m_blinkEye [0].Interval;
				}
			}
		}

		void Update()
		{
			BlinkControl ();
		}
		#endregion



		#region public methods -------------------------------------------------------------------
		/// <summary>
		/// 目の種類を変更.
		/// </summary>
		/// <param name="type">Type.</param>
		public void ChangeEyeType(EyeTypes type)
		{
			m_eyeInfo.Type = (int)type;

			// 座標を求める
			m_eyeInfo.Atlas.x = ((int)type / m_eyeInfo.VNum);
			m_eyeInfo.Atlas.y = ((int)type - (m_eyeInfo.Atlas.x * m_eyeInfo.VNum));
			m_eyeInfo.Atlas.x *= m_eyeInfo.Atlas.width;
			m_eyeInfo.Atlas.y *= m_eyeInfo.Atlas.height;

			// UV座標計算
			Vector2 offset;
			offset	= new Vector2(m_eyeInfo.Atlas.x / m_eyeInfo.TextureSize.x,  	1.0f-(m_eyeInfo.Atlas.y / m_eyeInfo.TextureSize.y));

			// 適用
			m_eyeInfo.Mat.SetTextureOffset("_MainTex", offset);
		}


		/// <summary>
		/// 口の種類を変更.
		/// </summary>
		/// <param name="type">Type.</param>
		public void ChangeMouthType(MouthTypes type)
		{
			m_mouthInfo.Type = (int)type;

			// 座標を求める
			m_mouthInfo.Atlas.x = ((int)type / m_mouthInfo.VNum);
			m_mouthInfo.Atlas.y = ((int)type - (m_mouthInfo.Atlas.x * m_mouthInfo.VNum));
			m_mouthInfo.Atlas.x *= m_mouthInfo.Atlas.width;
			m_mouthInfo.Atlas.y *= m_mouthInfo.Atlas.height;

			// UV座標計算
			Vector2 offset;
			offset	= new Vector2(m_mouthInfo.Atlas.x / m_mouthInfo.TextureSize.x,  	1.0f-(m_mouthInfo.Atlas.y / m_mouthInfo.TextureSize.y));

			// 適用
			m_mouthInfo.Mat.SetTextureOffset("_MainTex", offset);
		}

		#endregion



		#region private methods ------------------------------------------------------------------
		/// <summary>
		/// 瞬き制御
		/// </summary>
		private void BlinkControl()
		{
			float deltaTime = Time.deltaTime;
			for (int i = 0; i < (int)EYES.NUM; i++)
			{
				switch ((int)m_blinkEye [i].State)
				{
				case (int)BlinkState.READY:
					m_blinkEye [i].IntervalTimer += deltaTime;
					if (m_blinkEye [i].IntervalTimer >= m_blinkEye [i].Interval)
					{
						m_blinkEye [i].IntervalTimer = 0;
						m_blinkEye [i].State = BlinkState.CLOSING;
					}
					break;

				case (int)BlinkState.CLOSING:
					m_blinkEye [i].MotionTimer += deltaTime;
					if (m_blinkEye [i].MotionTimer >= m_blinkEye [i].Speed)
					{
						m_blinkEye [i].MotionTimer = 0;
						m_blinkEye [i].State = BlinkState.OPENING;
						if (i == (int)EYES.LEFT)
						{
							m_leftEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, 100);
						}
						else if (i == (int)EYES.RIGHT)
						{
							m_rightEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, 100);
						}
					}
					else
					{
						float blendValue = m_blinkEye [i].MotionTimer / m_blinkEye [i].Speed * 100f;
						if (i == (int)EYES.LEFT)
						{
							m_leftEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, blendValue);
						}
						else if (i == (int)EYES.RIGHT)
						{
							m_rightEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, blendValue);
						}
					}
					break;

				case (int)BlinkState.OPENING:
					m_blinkEye [i].MotionTimer += deltaTime;
					if (m_blinkEye [i].MotionTimer >= m_blinkEye [i].Speed)
					{
						m_blinkEye [i].MotionTimer = 0;
						m_blinkEye [i].State = BlinkState.READY;
						if (i == 0)
						{
							m_blinkEye [i].Interval = Random.Range (m_blinkEye [i].IntervalMin, m_blinkEye [i].IntervalMax);
						}
						else
						{
							// 両目で瞬きタイミングがずれないようにする
							m_blinkEye [i].Interval = m_blinkEye [0].Interval;
						}
						if (i == (int)EYES.LEFT)
						{
							m_leftEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, 0);
						}
						else if (i == (int)EYES.RIGHT)
						{
							m_rightEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, 0);
						}
					}
					else
					{
						float blendValue = 100 - m_blinkEye [i].MotionTimer / m_blinkEye [i].Speed * 100f;
						if (i == (int)EYES.LEFT)
						{
							m_leftEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, blendValue);
						}
						else if (i == (int)EYES.RIGHT)
						{
							m_rightEyeSkinnedMeshRenderer.SetBlendShapeWeight (0, blendValue);
						}
					}
					break;
				}
			}
		}
		#endregion
	}
}